@@ -1,8 +1,8 @@
 /* -*- c-basic-offset: 2; indent-tabs-mode: nil -*- */
 /*
-  Copyright(C) 2010 Tetsuro IKEDA
-  Copyright(C) 2010-2013 Kentoku SHIBA
-  Copyright(C) 2011-2017 Kouhei Sutou <kou@clear-code.com>
+  Copyright(C) 2010  Tetsuro IKEDA
+  Copyright(C) 2010-2013  Kentoku SHIBA
+  Copyright(C) 2011-2022  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -30,6 +30,11 @@
 #include <mrn_current_thread.hpp>
 
 #include <sql_table.h>
+#ifdef MRN_HAVE_UDF_METADATA
+#  include <mysql/components/my_service.h>
+#  include <mysql/components/services/udf_metadata.h>
+#  include <mysql/service_plugin_registry.h>
+#endif
 
 MRN_BEGIN_DECLS
 
@@ -42,11 +47,11 @@ struct CommandInfo
   grn_obj *db;
   bool use_shared_db;
   grn_obj command;
-  String result;
+  grn_obj result;
 };
 
 MRN_API mrn_bool mroonga_command_init(UDF_INIT *init, UDF_ARGS *args,
-                                     char *message)
+                                      char *message)
 {
   CommandInfo *info = NULL;
 
@@ -113,14 +118,48 @@ MRN_API mrn_bool mroonga_command_init(UDF_INIT *init, UDF_ARGS *args,
   init->maybe_null = 1;
   init->const_item = 0;
 
-  info = (CommandInfo *)mrn_my_malloc(sizeof(CommandInfo),
-                                      MYF(MY_WME | MY_ZEROFILL));
+  info = static_cast<CommandInfo *>(mrn_my_malloc(sizeof(CommandInfo),
+                                                  MYF(MY_WME | MY_ZEROFILL)));
   if (!info) {
     strcpy(message, "mroonga_command(): out of memory");
     goto error;
   }
 
   info->ctx = mrn_context_pool->pull();
+#ifdef MRN_HAVE_UDF_METADATA
+  {
+    const char *charset = nullptr;
+    grn_encoding encoding = GRN_CTX_GET_ENCODING(info->ctx);
+    if (encoding == GRN_ENC_DEFAULT) {
+      encoding = grn_get_default_encoding();
+    }
+    switch (GRN_CTX_GET_ENCODING(info->ctx)) {
+    case GRN_ENC_EUC_JP:
+      charset = "eucjpms";
+      break;
+    case GRN_ENC_UTF8:
+      charset = "utf8mb4";
+      break;
+    case GRN_ENC_SJIS:
+      charset = "cp932";
+      break;
+    case GRN_ENC_LATIN1:
+      charset = "latin1";
+      break;
+    case GRN_ENC_KOI8R:
+      charset = "koi8r";
+      break;
+    default:
+      break;
+    }
+    if (charset) {
+      auto *value = const_cast<char *>(charset);
+      my_service<SERVICE_TYPE(mysql_udf_metadata)>
+        service("mysql_udf_metadata", mysql_plugin_registry_acquire());
+      service->result_set(init, "charset", static_cast<void *>(value));
+    }
+  }
+#endif
   {
     const char *current_db_path = MRN_THD_DB_PATH(current_thd);
     const char *action;
@@ -155,8 +194,9 @@ MRN_API mrn_bool mroonga_command_init(UDF_INIT *init, UDF_ARGS *args,
     }
   }
   GRN_TEXT_INIT(&(info->command), 0);
+  GRN_TEXT_INIT(&(info->result), 0);
 
-  init->ptr = (char *)info;
+  init->ptr = reinterpret_cast<char *>(info);
 
   return false;
 
@@ -251,7 +291,7 @@ MRN_API char *mroonga_command(UDF_INIT *init, UDF_ARGS *args, char *result,
     goto error;
   }
 
-  info->result.length(0);
+  GRN_BULK_REWIND(&(info->result));
   do {
     char *buffer;
     unsigned int buffer_length;
@@ -261,16 +301,12 @@ MRN_API char *mroonga_command(UDF_INIT *init, UDF_ARGS *args, char *result,
       goto error;
     }
     if (buffer_length > 0) {
-      if (info->result.reserve(buffer_length)) {
-        my_error(ER_OUT_OF_RESOURCES, MYF(0), HA_ERR_OUT_OF_MEM);
-        goto error;
-      }
-      info->result.q_append(buffer, buffer_length);
+      GRN_TEXT_PUT(ctx, &(info->result), buffer, buffer_length);
     }
   } while (flags & GRN_CTX_MORE);
 
-  *length = info->result.length();
-  return (char *)(info->result.ptr());
+  *length = GRN_TEXT_LEN(&(info->result));
+  return GRN_TEXT_VALUE(&(info->result));
 
 error:
   *error = 1;
@@ -281,12 +317,12 @@ MRN_API void mroonga_command_deinit(UDF_INIT *init)
 {
   CommandInfo *info = (CommandInfo *)init->ptr;
   if (info) {
+    GRN_OBJ_FIN(info->ctx, &(info->result));
     GRN_OBJ_FIN(info->ctx, &(info->command));
     if (!info->use_shared_db) {
       grn_obj_close(info->ctx, info->db);
     }
     mrn_context_pool->release(info->ctx);
-    MRN_STRING_FREE(info->result);
     my_free(info);
   }
 }
